<?php
/**
 * WooCommerce AvaTax
 *
 * This source file is subject to the GNU General Public License v3.0
 * that is bundled with this package in the file license.txt.
 * It is also available through the world-wide-web at this URL:
 * http://www.gnu.org/licenses/gpl-3.0.html
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@skyverge.com so we can send you a copy immediately.
 *
 * DISCLAIMER
 *
 * Do not edit or add to this file if you wish to upgrade WooCommerce AvaTax to newer
 * versions in the future. If you wish to customize WooCommerce AvaTax for your
 * needs please refer to http://docs.woocommerce.com/document/woocommerce-avatax/
 *
 * @author    SkyVerge
 * @copyright Copyright (c) 2016-2022, SkyVerge, Inc.
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU General Public License v3.0
 */

namespace SkyVerge\WooCommerce\AvaTax\API\Models;

use SkyVerge\WooCommerce\PluginFramework\v5_10_14 as Framework;

defined( 'ABSPATH' ) or exit;

/**
 * An object representation of a classification model to be used by the Avalara API.
 *
 * @since 1.13.0
 */
class HS_Classification_Model {


	/** @var string unique ID generated by Avalara classification system */
	protected $id;

	/** @var HS_Item_Model product details */
	protected $item;

	/** @var string required classification activity for a request */
	protected $activity;

	/** @var string country of destination code */
	protected $country_of_destination;

	/** @var string HS classification result */
	protected $hs_code;

	/** @var string requested HS code status */
	protected $status;

	/** @var string contains a detailed description of an issue found */
	protected $resolution;

	/** @var string rating for classification purposes when the information provided is insufficient */
	protected $confidence;

	/** @var string detailed description of the confidence rating above, when applicable */
	protected $confidence_description;


	/**
	 * Gets the classification identifier.
	 *
	 * @since 1.13.0
	 *
	 * @return string
	 */
	public function get_id() : string {

		return $this->id;
	}


	/**
	 * Sets an id for the classification.
	 *
	 * @since 1.13.0
	 *
	 * @param string $value
	 * @return self
	 */
	public function set_id( string $value ) : HS_Classification_Model {

		$this->id = $value;

		return $this;
	}


	/**
	 * Gets the item.
	 *
	 * @since 1.13.0
	 *
	 * @return HS_Item_Model
	 */
	public function get_item() : HS_Item_Model {

		return $this->item;
	}


	/**
	 * Sets the item.
	 *
	 * @since 1.13.0
	 *
	 * @param HS_Item_Model $value
	 * @return $this
	 */
	public function set_item( HS_Item_Model $value ) : HS_Classification_Model {

		$this->item = $value;

		return $this;
	}


	/**
	 * Gets the required classification activity.
	 *
	 * @return string
	 */
	public function get_activity() : string {

		return $this->activity;
	}


	/**
	 * Sets the classification activity.
	 *
	 * @since 1.13.0
	 *
	 * @param string $value
	 * @return self
	 */
	public function set_activity( string $value ) : HS_Classification_Model {

		$this->activity = $value;

		return $this;
	}


	/**
	 * Gets the country of destination
	 *
	 * @since 1.13.0
	 *
	 * @return string
	 */
	public function get_country_of_destination() : string {

		return $this->country_of_destination;
	}


	/**
	 * Sets the country of destination.
	 *
	 * @since 1.13.0
	 *
	 * @param string $value
	 * @return self
	 */
	public function set_country_of_destination( string $value ) : HS_Classification_Model {

		$this->country_of_destination = $value;

		return $this;
	}


	/**
	 * Gets the HS code.
	 *
	 * @since 1.13.0
	 *
	 * @return string
	 */
	public function get_hs_code() : string {

		return $this->hs_code;
	}


	/**
	 * Sets the HS code.
	 *
	 * @since 1.13.0
	 *
	 * @param string $value
	 * @return self
	 */
	public function set_hs_code( string $value ) : HS_Classification_Model {

		$this->hs_code = $value;

		return $this;
	}


	/**
	 * Gets the status.
	 *
	 * @since 1.13.0
	 *
	 * @return string
	 */
	public function get_status() : string {

		return $this->status;
	}


	/**
	 * Sets the status.
	 *
	 * @since 1.13.0
	 *
	 * @param string $value
	 * @return self
	 */
	public function set_status( string $value ) : HS_Classification_Model {

		$this->status = $value;

		return $this;
	}


	/**
	 * Gets the resolution.
	 *
	 * @since 1.13.0
	 *
	 * @return string
	 */
	public function get_resolution() : string {

		return $this->resolution;
	}


	/**
	 * Sets the resolution.
	 *
	 * @since 1.13.0
	 *
	 * @param string $value
	 * @return self
	 */
	public function set_resolution( string $value ) : HS_Classification_Model {

		$this->resolution = $value;

		return $this;
	}


	/**
	 * Gets the confidence value.
	 *
	 * @since 1.13.0
	 *
	 * @return string
	 */
	public function get_confidence() : string {

		return $this->confidence;
	}


	/**
	 * Sets the confidence.
	 *
	 * @since 1.13.0
	 *
	 * @param string $value
	 * @return self
	 */
	public function set_confidence( string $value ) : HS_Classification_Model {

		$this->confidence = $value;

		return $this;
	}


	/**
	 * Gets the confidence description.
	 *
	 * @since 1.13.0
	 *
	 * @return string
	 */
	public function get_confidence_description() : string {

		return $this->confidence_description;
	}


	/**
	 * Sets the confidence description.
	 *
	 * @since 1.13.0
	 *
	 * @param string $value
	 * @return self
	 */
	public function set_confidence_description( string $value ) : HS_Classification_Model {

		$this->confidence_description = $value;

		return $this;
	}


}
